
#include <windows.h>
#include <iostream>
//#include <math.h>
#include "cv.h"
#include "cvaux.h"
#include "highgui.h"

using namespace cv;
using namespace std;

#include "common.h"
#include "auto_eye_tracker.h"
#include "queue.h"

#include "compositor.h"
#define TEXT_OUTPUT_ENABLE	1
#define VIDEO_FILE_ENABLE	0

static DWORD WINAPI thread(LPVOID pv_params)
{
	compositor	*ctxt = (compositor*)pv_params;

	ctxt->process();

	return 0;
}

compositor::compositor(char* session_name, queue_t *in_scene, queue_t *in_eye_coords)
{
	char filename[100];

	this->in_scene = in_scene;
	this->in_eye_coords = in_eye_coords;
	this->mode = COMP_CALIBRATE;
	this->last_eye_point = Point(-1, -1);
	this->last_scene_point = Point(-1, -1);
	this->tracked_eye = Point(-1,-1);
	
	this->video_file = NULL;
	this->text_file = NULL;

	strcpy(this->session_name, session_name);

#if TEXT_OUTPUT_ENABLE
	//Open the text file
	sprintf(filename, "%s.txt", session_name);
	this->text_file = fopen(filename, "w");
	if( NULL == this->text_file)
	{
		cout <<"Error: Text file " << filename << ".txt could not be opened\n";
		return;
	}
#endif

	this->keep_running = true;
	this->h_thread = CreateThread(	NULL,                   // default security attributes
									0,                      // use default stack size  
									thread,					// thread function name
									this,					// argument to thread function 
									0,                      // use default creation flags 
									&this->thread_id);		// returns the thread identifier 

	if(NULL == this->h_thread)
	{
		cout << "Error in creating Hough Transforms thread\n";
		return;
	}
}


compositor::~compositor(void)
{
#if TEXT_OUTPUT_ENABLE
	if( NULL != this->text_file)
		fclose(this->text_file);
	
#endif

#if VIDEO_FILE_ENABLE
	if( NULL != this->video_file)
	{
		delete this->video_file;
	}
#endif

	//Signal the thread to close.
	this->keep_running = false;
	
	// Wait for it to finish.
	WaitForSingleObject(this->h_thread, INFINITE);

	//That is it!
	cout << "Hough Transform closed.\n";
}


status_t compositor::set_mode(compositor_modes_t mode)
{
	this->mode = mode;
	if( mode == COMP_TRACK)
	{
		cout << "Compositor: Tracking mode\n";
	}
	else
	{
		cout << "Compositor: calibration mode\n";
	}
	Beep(500, 400);
	return SUCCESS;
}


status_t compositor::process()
{
	//For now, this process simply tries to find the 'bright spot'.
	status_t	error;
	Mat		*ps_frame; //Input frame.

	Point	*ps_eye;
	int		alpha = DEFAULT_COMPOSITOR_ALPHA;
	bool	trackbar_init = false;

	int gaze_error;
	
	while( this->keep_running)
	{
		// Read the input frame.
		do
		{
			ps_frame = NULL;
			error = this->in_scene->dequeue((void**)&ps_frame, 100);
		}while(TIMEOUT == error && this->keep_running);

		if(SUCCESS != error || ps_frame == NULL) 
		{
			cout << "Error in reading input frame\n" << error ;
			continue;
		}

		//Read the eye coordinates.
		do
		{
			ps_eye = NULL;
			error = this->in_eye_coords->dequeue((void**)&ps_eye, 100);
		}while(TIMEOUT == error && this->keep_running);

		if(SUCCESS != error || NULL == ps_eye) 
		{
			cout << "Compositor: Error in reading eye coordinates\n" << error ;
			continue;
		}

		//find the brightest point on the screen.
		Point	bright_point(-1,-1),  min_point(-1,-1);
		Mat		s_mask = Mat::ones(ps_frame->rows, ps_frame->cols, CV_8UC1);
		double  max, min;

		//Calculate the position of the bright spot. 
		//The assumption is that the subject is looking at it!
		//It is supposed to be a flashlight pointing at the camera.
		max = 0;
		minMaxLoc(*ps_frame, &min, &max, &min_point, &bright_point, s_mask );

		circle( *ps_frame, bright_point, 5, Scalar(0,0,0), -1);
		circle( *ps_frame, bright_point, 10, Scalar(0,0,0), 2);

		//Check the state, and act accordingly.
		if( this->mode == COMP_CALIBRATE)
		{
			this->calibrate(ps_frame, ps_eye, bright_point);
		}
		else if(this->mode == COMP_TRACK)
		{
			Point *gaze_point = NULL;
			int colour = 255;

			error = this->track(ps_frame, ps_eye, &gaze_point, alpha);


			//have a trackbar for alpha
			if( !trackbar_init)
			{
				createTrackbar( "alpha",
								"gaze_point",
								&alpha, 
								99,
								NULL,
								NULL);
				trackbar_init = true;
			}

			alpha = getTrackbarPos("alpha", "gaze_point");



			//Write to file.
#if VIDEO_FILE_ENABLE	
			if( NULL == this->video_file)
			{
				char filename[500];
				Size framesize(ps_frame->cols, ps_frame->rows);
				sprintf(filename, "%s.avi", this->session_name);
				printf("Opening file %s\n", filename);
				this->video_file = new VideoWriter(filename, CV_FOURCC('D','I','V','X'), 30, framesize , true);

				if( NULL == this->video_file)
					cout << "Comp: videofile couldn't be opened\n";


			}

				//Convert to 3-channel.
				Mat dummy;
				cvtColor(*ps_frame, dummy, CV_GRAY2BGR);
				if( this->video_file)
					*(this->video_file) << dummy;
#endif


#if TEXT_OUTPUT_ENABLE
			{
				//Calculate error as gaze_point-brightestpoint
				float xdiff = (gaze_point->x - bright_point.x );
				float ydiff = (gaze_point->y - bright_point.y );

				SYSTEMTIME	s_time;
				GetLocalTime(&s_time);

				gaze_error = (int) sqrt( xdiff * xdiff + ydiff*ydiff);

				fprintf(this->text_file, "%02d:%02d:%02d:%03d: x = %03d, y = %03d; bright-point diff = %d\n", 
						s_time.wHour, s_time.wMinute, s_time.wSecond, s_time.wMilliseconds,
						gaze_point->x, gaze_point->y, gaze_error);
			}
#endif

			//Draw the bulls-eye!
			circle( *ps_frame, *gaze_point, 5, Scalar(255,255,255), -1);
			circle( *ps_frame, *gaze_point, 15, Scalar(255,255,255), 2);
			if( SUCCESS ==  error)
			{
				circle( *ps_frame, *gaze_point, 25, Scalar(255,255,255), 2);
			}

			delete gaze_point;
		}

		imshow("Final output", *ps_frame);
		waitKey(1);


		delete ps_frame;
	}

	return SUCCESS;
}

//Tell the compositor to calibrate using current data.

status_t compositor::calibrate_on_next_frame()
{
	this->calibrate_now = true;
	return SUCCESS;
}


/*****************
 Calibration of the eye tracker
 *****************/
status_t compositor::calibrate(Mat *ps_frame, Point *ps_eye, Point bright_point)
{
	bool calibrated = false;

	if(this->calibrate_now == false)
	{
		return SUCCESS;
	} 

	//Go ahead and calibrate.
	this->calibrate_now = false;

	//If no previous point has been received yet.
	if( -1 == this->last_scene_point.x )
	{
		this->last_scene_point.x = bright_point.x;
		this->last_scene_point.y = bright_point.y;
		this->last_eye_point.x = ps_eye->x;
		this->last_eye_point.y = ps_eye->y;
	}
	else
	{

		/***********************************************************************
		Calibrate X-axis.
		Calibrate only if:
			- Current scene point is frame-size/4 away from previous point
			- AND the last eye point is not the same as the current eye point.
				(The second condition will cause a divide-by-zero)
				
		************************************************************************/
		if( abs(bright_point.x - this->last_scene_point.x) > ps_frame->cols/4 && 
			(ps_eye->x - this->last_eye_point.x) != 0)
		{
			calibrated = true;

			this->eye_offset.x = this->last_eye_point.x;
			this->scene_offset.x = this->last_scene_point.x;
			this->scale.x = ((float)bright_point.x - (float)this->last_scene_point.x) / 
							((float)ps_eye->x - (float)this->last_eye_point.x);
			


			printf("X:Last eye point = %d, %d; last scene point = %d, %d \n",
					this->last_eye_point.x,this->last_eye_point.y, 
					this->last_scene_point.x, this->last_scene_point.y );
				
			printf("X:Current eye point = %d, %d; current scene point = %d, %d \n",
					ps_eye->x,ps_eye->y, 
					bright_point.x, bright_point.y );

			printf("Calib: eyeoff.x = %d, xscale = %3.2f, scene_off.x = %d\n",
					this->eye_offset.x, this->scale.x, this->scene_offset.x);
						//Update the last points.
			this->last_eye_point.x = ps_eye->x;
			this->last_scene_point.x = bright_point.x;

		}

		/***********************************************************************
		Calibrate Y-axis.
		Calibrate only if:
			- Current scene point is frame-size/4 away from previous point
			- AND the last eye point is not the same as the current eye point.
				(The second condition will cause a divide-by-zero)
				
		************************************************************************/
		if( (ps_eye->y - this->last_eye_point.y) != 0 && 
			 abs(bright_point.y - this->last_scene_point.y) > ps_frame->rows/4 )
		{
			calibrated = true;
			
			this->eye_offset.y = this->last_eye_point.y;
			this->scene_offset.y = this->last_scene_point.y;
			this->scale.y = ((float)bright_point.y - (float)this->last_scene_point.y) / 
							((float)ps_eye->y - (float)this->last_eye_point.y);

				
			printf("Y:Last eye point = %d, %d; last scene point = %d, %d \n",
					this->last_eye_point.x,this->last_eye_point.y, 
					this->last_scene_point.x, this->last_scene_point.y );
				
			printf("Y:Current eye point = %d, %d; current scene point = %d, %d \n",
					ps_eye->x,ps_eye->y, 
					bright_point.x, bright_point.y );

			printf("Calib: eyeoff.y = %d, yscale = %3.2f, scene_off = %d\n",
					this->eye_offset.y, this->scale.y, this->scene_offset.y);
			
			this->last_eye_point.y = ps_eye->y;
			this->last_scene_point.y = bright_point.y;
		}
	}

	if( calibrated )
	{
		Beep(1000,200);
		Beep(500, 200);
	}

	return SUCCESS;
}

/*****************
 Using previously calculated scales, convert from eye coordinates to scene coorinates.

 Xscene = (x-xe)(xs' - xs)/(xe'-xe) + xs where
			 xe->xs
			 xe'->xs'

 *****************/

status_t compositor::track(Mat *ps_frame, Point *ps_eye, Point **pps_gaze, int alpha)
{
	Point *gaze = new Point;
	float f_alpha = alpha/100;
	status_t error = SUCCESS;

	gaze->x = (ps_eye->x - this->eye_offset.x)*(this->scale.x) + this->scene_offset.x;
	gaze->y = ( ps_eye->y - this->eye_offset.y)*(this->scale.y) + this->scene_offset.y;

	if( gaze->x >= ps_frame->cols )
	{
		gaze->x = ps_frame->cols - 5;
		error = OUT_OF_RANGE;
	}

	if( gaze->x < 0 )
	{
		gaze->x =  5;
		error = OUT_OF_RANGE;
	}

	if( gaze->y >= ps_frame->rows )
	{
		gaze->y = ps_frame->rows - 5;
		error = OUT_OF_RANGE;
	}

	if( gaze->y < 0 )
	{
		gaze->y = 5;
		error = OUT_OF_RANGE;
	}

	//Smoothen
	gaze->x = this->tracked_eye.x * f_alpha + (1-f_alpha)*(gaze->x);
	gaze->y = this->tracked_eye.y * f_alpha + (1-f_alpha)*(gaze->y);
	this->tracked_eye = *gaze;

	*pps_gaze = gaze;


	return SUCCESS;
}


