/*************************************************************
Class to identify the rough centre of the eye. Specifies
a rough region using the ROI parameters.
*************************************************************/
#include <windows.h>
#include <iostream>
using namespace std;

#include "cv.h"
#include "cvaux.h"
#include "highgui.h"
using namespace cv;

#include "common.h"
#include "queue.h"
#include "auto_eye_tracker.h"
#include "eye_analogue.h"


//This function is defined in find_eye_analogues.cpp
status_t find_eye_analogue(	Mat *ps_input, 
						Mat	*ps_eye_analogue,
						int  h, 
						int	 w);


//Thread function. Strictly a dummy placeholder!
static DWORD WINAPI thread(LPVOID pv_params)
{
	eye_analogue	*ctxt = (eye_analogue*)pv_params;

	ctxt->process();

	return 0;
}


eye_analogue::eye_analogue(int eye_window_h, 
						   int eye_window_w,
						   queue_t* eye_queue,
						   queue_t* scene_queue,
						   queue_t* out_scene,
						   queue_t* out_original,
						   queue_t* out_processed)
						   
{
	this->h = eye_window_h;
	this->w = eye_window_w;

	this->q_scene_in = scene_queue;
	this->input_q    = eye_queue;

	this->out_original = out_original;
	this->out_processed = out_processed;

	this->q_scene_out = out_scene;

	this->keep_running = true;

	//Create the thread.
	this->h_thread = CreateThread(	NULL,                   // default security attributes
									0,                      // use default stack size  
									thread,					// thread function name
									this,					// argument to thread function 
									0,                      // use default creation flags 
									&this->thread_id);		// returns the thread identifier 

	if(NULL == this->h_thread)
	{
		cout << "Error in creating Eye-analogues thread\n";
		return;
	}
}


eye_analogue::~eye_analogue(void)
{

	//Signal the thread to close.
	this->keep_running = false;
	
	// Wait for it to finish.
	WaitForSingleObject(this->h_thread, INFINITE);

	//That is it!
	cout << "Eye-analogye closed.\n";
}


//Actual thread function.
status_t eye_analogue::process()
{
	Mat s_resized_frame;
	
	Mat *ps_input_frame = NULL;
	Mat *ps_scene = NULL;
	Mat *ps_processed_frame = NULL;
	status_t	error;

	int newh;
	int neww;

	int start;

	while( this->keep_running)
	{
		//Get an input frame.
		GETBUF(input_q, ps_input_frame, 100)
		if( SUCCESS != error || NULL == ps_input_frame)
		{
			cout << "Eye analogue: error in getting input frame, " << error << endl;
			continue;
		}

		GETBUF(q_scene_in, ps_scene, 100)
		if(SUCCESS != error || NULL == ps_scene)
		{
			cout << "Eye analogue: error in getting scene frame, " << error << endl;
			delete ps_input_frame;
			continue;
		}

		//Resize this frame.
		resize(*ps_input_frame, s_resized_frame, Size(RESIZE_OUT_W, RESIZE_OUT_H));

		newh = (h * s_resized_frame.rows)/ps_input_frame->rows;
		neww = (w * s_resized_frame.cols)/ps_input_frame->cols;
		
		//Run the find-analogues algorithm on the resized image.
		ps_processed_frame = new Mat;
		*ps_processed_frame = Mat::zeros(s_resized_frame.rows, s_resized_frame.cols, CV_8UC1);
		
		start = clock();
		
		error = find_eye_analogue(&s_resized_frame, ps_processed_frame, newh, neww);
		if( SUCCESS != error)
		{
			cout << "Error: find_eye_analogue returned \n" << error << endl;
			delete ps_input_frame;
			delete ps_scene;
			continue;
		}

		GaussianBlur(*ps_processed_frame, *ps_processed_frame, Size(3,3), 1.5, 1.5);

		//printf("eye-analogue took %d ms\n", clock() - start);
		imshow("analogue", *ps_processed_frame);
		waitKey(1);

		//push the two frames out.
		if( this->out_original)
		{
			error = this->out_original->enqueue(ps_input_frame, 100);
			if( SUCCESS != error)
			{
				cout << "Eye-analogues: Unable to push out input frame " << error << endl;
				delete ps_input_frame;
				delete ps_processed_frame;
				continue;
			}
		}
		else
		{
			delete ps_input_frame;
		}
		
		//Push out the processed frame.
		if( this->out_processed)
		{
			error = this->out_processed->enqueue(ps_processed_frame, 100);
			if( SUCCESS != error)
			{
				if( TIMEOUT != error)
					cout << "Eye-analogues: Unable to push out output frame "<< error << endl;
				delete ps_processed_frame;
			}
		}
		else
		{
			delete ps_processed_frame;
		}

		//Push out the scene frame.
		if( this->q_scene_out)
		{
			error = this->q_scene_out->enqueue(ps_scene, 100);
			if( SUCCESS != error)
			{
				if( TIMEOUT != error)
					cout << "Eye-analogues: Unable to push out scene frame "<< error << endl;
				delete ps_scene;
			}
		}
	}

	return SUCCESS;
}