/********************************************************************
find_eye_analogues.c Compute possible Eye-Windows from a face image.
v1.0 - Initial version.											
********************************************************************/

#include "cv.h"
#include "cvaux.h"
#include "highgui.h"

using namespace cv;

#include "common.h"
#include "auto_eye_tracker.h"


#define PROFILING_ON

#ifdef PROFILING_ON
	#define TIMER_START	start = clock()
	#define TIMER_END	end = clock()
#else
	#define	TIMER_START
	#define TIMER_END
#endif

//Number of regions read around each pixel.
#define NUM_REGIONS				8

//Threshold number of regions that must satisfy the test.
#define MEAN_HITS_THRESHOLD		6



static void derive_from_last_line(Mat *ps_input, 
						   unsigned int* line_start_sum, 
						   unsigned int* prev_sum, 
						   int x, 
						   int y, 
						   int half_h, 
						   int half_w);


static void derive_from_prev_pixel_transposed(Mat *ps_input, 
									   Mat *ps_transpose,
									   unsigned int* prev_sum, 
									   int x, 
									   int y, 
									   int half_h, 
									   int half_w,
									   int *times);
static void fill_pixel_sizes(int *num_pixels, int half_h, int half_w);
/**************************************************************************************
find_eye_analogue: Calculates possible eye-analogue pixels using mean values.

Inputs: ps_input: Input frame with an eye image.
		ps_eye_analogue: Output frame. It is assumed that this frame has been
						 allocated; it is overwritten.
		h : Height of eye window. Should be an even number of pixels.
		w : Width of eye window. Should be an even number of pixels.

Note: This is the 'smart' version of this function. It does not calculate the means 
	  of each region by brute force; it calculates them from the previous regions.
	  This requires the storage of the previous values.


Regions: The whole rectangle is wxh in size
	---------------------------------
	|				|				|
	|	0		   1|		2		|
	|-------3-------x-------4--------
	|				|				|
	|	5		   6|		7		|
	---------------------------------
**************************************************************************************/

status_t find_eye_analogue(	Mat *ps_input,
						Mat	*ps_eye_analogue,
						int  h, 
						int	 w)
{
	unsigned int line_start_sum[NUM_REGIONS];
	unsigned int prev_sum[NUM_REGIONS];

	int x, y,i;
	int num_hits;
	int cur_pixel;

	int num_pixels[NUM_REGIONS];
	int times[NUM_REGIONS];

	int half_w = w/2;
	int half_h = h/2;

	int nrows = ps_input->rows;
	int ncols = ps_input->cols;

	int xlow, xhigh, ylow, yhigh;

	int start = clock();
	int end;

	Mat s_transpose;

	fill_pixel_sizes(num_pixels, half_h, half_w);

	xlow  = half_w+1;
	xhigh = ncols - half_w-1;
	ylow  = half_h+1;
	yhigh = nrows - half_h-1;

	//Validate the input.
	{
		if(ps_input->empty())
			return EMPTY_FRAME;
		if( ncols > MAX_COLS_POSSIBLE || ncols < 0 || xlow > xhigh)
			return INVALID_COLS;
		if( nrows > MAX_ROWS_POSSIBLE || nrows < 0 || ylow > yhigh)
			return INVALID_ROWS;
	}

	//Create a transponse matrix.
	transpose(*ps_input, s_transpose);

	for( y = ylow; y < yhigh; y++)
	{
		for(x = xlow ; x < xhigh; x++)
		{
			if( xlow == x)
			{
				//This is the first pixel in the line.
				if(ylow == y)
				{
					//This is the first pixel in the frame. No history to bank upon.
					start = clock();
					//Calculate region by region
					prev_sum[0] = find_sum(ps_input, x-half_w,	y-half_h, half_h, half_w);
					
					prev_sum[1] = find_sum(ps_input, x,			y-half_h, half_h, 1);

					prev_sum[2] = find_sum(ps_input, x+1,		y-half_h, half_h, half_w);

					prev_sum[3] = find_sum(ps_input, x-half_w,  y,		  1,	  half_w);

					prev_sum[4] = find_sum(ps_input, x+1,		y,		  1,	  half_w);

					prev_sum[5] = find_sum(ps_input, x-half_w,	y+1,	  half_h, half_w);

					prev_sum[6] = find_sum(ps_input, x,			y+1,	  half_h, 1);

					prev_sum[7] = find_sum(ps_input, x+1,		y+1,	  half_h, half_w);

					end = clock();
					//DBG "first pixel: took %d ms\n", end - start);

				}
				else
				{
					/*Derive the values from the previous line's start element; replace 
					that with the current values.*/
					start = clock();
					derive_from_last_line(ps_input, line_start_sum, prev_sum, x, y, half_h, half_w);
					end = clock();
				}

				//Copy these into the line-start sum as well.
				for(i=0; i<NUM_REGIONS; i++)
				{
					line_start_sum[i] = prev_sum[i];
				}
			}
			else
			{
				//Smart mean - derive the sum from the preceeding pixel's value.

				derive_from_prev_pixel_transposed(ps_input, &s_transpose, prev_sum, x, y, half_h, half_w, times);

			}

			//At this point, prev_point[i] contains the sums of values in the relevant regions.
			//Compare the means - not the sums.
			num_hits = 0;
			cur_pixel = PIXEL(ps_input, x, y); //ps_input->u[y][x];

			for(i = 0; i< NUM_REGIONS; i++)
			{
				if ( cur_pixel <  0.9 * (prev_sum[i]/num_pixels[i]) )
				{
					num_hits++;
				}
			}
			
			PIXEL(ps_eye_analogue, x, y) = (num_hits >= MEAN_HITS_THRESHOLD)? 0xFF: 0x00;
		}
	}

	return SUCCESS;
}


//Calculate mean around a pixel.
unsigned int find_sum( Mat *ps_input, 
			  			int x_pos, 
						int y_pos, 
						int	h, 
						int	w)
{
	//Cut out the matrix.
	Mat rows = ps_input->rowRange(y_pos, y_pos+h);
	Mat small = rows.colRange(x_pos, x_pos+w);
	Scalar sum2 = sum(small);
	unsigned int sum1 = sum2.val[0];
	
	return sum1;
}

//Derive the current values from the values in the last line.
static void derive_from_last_line(Mat *ps_input, 
						   unsigned int* line_start_sum, 
						   unsigned int* prev_sum, 
						   int x, 
						   int y, 
						   int half_h, 
						   int half_w)
{
	//Sum of top and bottom-most lines.
	int sum_top;
	int sum_bot;

	int prev_x = x;
	int prev_y = y-1;
	int region = 0;

	//Region 0
	sum_top = find_sum(ps_input, prev_x-half_w, prev_y-half_h, 1, half_w);
	sum_bot = find_sum(ps_input, prev_x-half_w, prev_y,		   1, half_w);
	prev_sum[region] = line_start_sum[region] - sum_top + sum_bot;
	region++;
	
	//Region 1
	sum_top = PIXEL(ps_input, x, (prev_y-half_h));
	sum_bot = PIXEL(ps_input, x, prev_y);
	prev_sum[region] = line_start_sum[region] - sum_top + sum_bot;
	region++;

	//Region 2
	sum_top = find_sum(ps_input, prev_x, prev_y-half_h, 1, half_w);
	sum_bot = find_sum(ps_input, prev_x, prev_y,		1, half_w);
	prev_sum[region] = line_start_sum[region] - sum_top + sum_bot;
	region++;
	
	//Region 3
	prev_sum[region] = find_sum(ps_input, x-half_h, y, 1, half_w);
	region++;

	//Region 4
	prev_sum[region] = find_sum(ps_input, x+1, y, 1, half_w);
	region++;

	//Region 5
	sum_top = find_sum(ps_input, x-half_w, y, 1, half_w);
	sum_bot = find_sum(ps_input, x-half_w, y+half_h, 1, half_w);
	prev_sum[region] = line_start_sum[region] - sum_top + sum_bot;
	region++;

	//Region 6
	sum_top = PIXEL(ps_input, x, y);
	sum_bot = PIXEL(ps_input, x, y+half_h);
	prev_sum[region] = line_start_sum[region] - sum_top + sum_bot;
	region++;

	//Region 7
	sum_top = find_sum(ps_input, x+1, y, 1, half_w);
	sum_bot = find_sum(ps_input, x+1, y+half_h, 1, half_w);
	prev_sum[region] = line_start_sum[region] - sum_top + sum_bot;
	region++;
}


#ifdef PROFILING_ON
	#define STORE_TIME times[region] = end - start;
#else
	#define STORE_TIME
#endif



/****************************************************************
Function to derive the pixel values using information about
previous pixels' sum values.
****************************************************************/
static void derive_from_prev_pixel_transposed(Mat *ps_input, 
									   Mat *ps_transpose,
									   unsigned int* prev_sum, 
									   int x, 
									   int y, 
									   int half_h, 
									   int half_w,
									   int *times)
{
	int remove;
	int add;
	int region = 0;
	int prev_x = x-1;
	int prev_y = y;
	int start, end;

	//region 0
	TIMER_START;
	remove = find_sum(ps_transpose, prev_y-half_h, prev_x-half_w, 1, half_h);
	TIMER_END;
	STORE_TIME;
	add = find_sum(ps_transpose, prev_y-half_h, prev_x, 1, half_h);
	prev_sum[region] = prev_sum[region] - remove + add;
	region++;

	//Region 1
	TIMER_START;
	prev_sum[region] = find_sum(ps_transpose, y-half_h, x, 1, half_h);
	TIMER_END;
	STORE_TIME;
	region++;
	

	TIMER_START;
	//Region 2
	remove = prev_sum[1]; //find_sum(ps_input, x, y-half_h, half_h, 1);
	TIMER_END;
	STORE_TIME;
	add = find_sum(ps_transpose, prev_y-half_h,prev_x+half_w, 1, half_h);
	prev_sum[region] = prev_sum[region] - remove + add;
	region++;
	

	TIMER_START;
	//Region 3
	prev_sum[region] = find_sum(ps_input, x-half_h, y, 1, half_w);
	TIMER_END;
	STORE_TIME;
	region++;

	TIMER_START;
	//Region 4
	prev_sum[region] = find_sum(ps_input, x+1, y, 1, half_w);
	TIMER_END;
	STORE_TIME;
	region++;

	TIMER_START;
	//Region 5
	remove = find_sum(ps_transpose, prev_y, prev_x-half_h, 1,half_h);
	TIMER_END;
	STORE_TIME;
	add = find_sum(ps_transpose, prev_y, prev_x, 1, half_h);
	prev_sum[region] = prev_sum[region] - remove + add;
	region++;

	TIMER_START;
	//Region 6
	prev_sum[region] = find_sum(ps_transpose, y, x, 1, half_h);
	TIMER_END;
	STORE_TIME;
	region++;


	TIMER_START;
	//Region 7
	remove = prev_sum[6];//find_sum(ps_input, x, y, half_h, 1);
	TIMER_END;
	STORE_TIME;
	add = find_sum(ps_transpose, prev_y, prev_x+half_w, 1, half_h);
	prev_sum[region] = prev_sum[region] - remove + add;
	region++;
}


//Smalll helper funciton to derive region sizes from h/w values.
static void fill_pixel_sizes(int *num_pixels, int half_h, int half_w)
{
	num_pixels[0] = half_h * half_w;
	num_pixels[1] = half_h;
	num_pixels[2] = half_h * half_w;
	num_pixels[3] = half_w;
	num_pixels[4] = half_w;
	num_pixels[5] = half_h * half_w;
	num_pixels[6] = half_h;
	num_pixels[7] = half_h * half_w;
}