
#include <windows.h>
#include <iostream>
#include "cv.h"
#include "cvaux.h"
#include "highgui.h"

using namespace cv;
using namespace std;

#include "common.h"
#include "auto_eye_tracker.h"
#include "queue.h"
#include "hough_transform.h"


static DWORD WINAPI thread(LPVOID pv_params)
{
	hough_transform	*ctxt = (hough_transform*)pv_params;

	ctxt->process();

	return 0;
}

hough_transform::hough_transform(int h, int w, queue_t* in_video, queue_t* in_approx_pos, queue_t *in_scene,
								 queue_t* out_coords, queue_t *out_scene)
{
	this->h = h;
	this->w = w;
	this->in_video = in_video;
	this->in_approx_pos = in_approx_pos;
	
	this->out_coords = out_coords;
	this->last_centre = Point(-1,-1);
	this->in_scene = in_scene;
	this->out_scene = out_scene;

	this->keep_running = true;
	this->h_thread = CreateThread(	NULL,                   // default security attributes
									0,                      // use default stack size  
									thread,					// thread function name
									this,					// argument to thread function 
									0,                      // use default creation flags 
									&this->thread_id);		// returns the thread identifier 

	if(NULL == this->h_thread)
	{
		cout << "Error in creating Hough Transforms thread\n";
		return;
	}

}


hough_transform::~hough_transform(void)
{
	//Signal the thread to close.
	this->keep_running = false;
	
	// Wait for it to finish.
	WaitForSingleObject(this->h_thread, INFINITE);

	//That is it!
	cout << "Hough Transform closed.\n";
}

//Thread function.
status_t hough_transform::process()
{
	status_t	error;

	Mat		*ps_frame; //Input frame.
	Point	*ps_approx; //Approximate eye location.

	int newh, neww;
	int hough_acc_thresh = 10;
	int min_radius = 48;
	int canny_param = 37;

	int dummy = 0;
	
	int thresh = DEFAULT_EYE_THRESH;

	Point eyew_ul;
	Point eyew_br;
	Point centre;//Centre of the eye

	bool centre_found;

	int alpha = ALPHA;
	

	while( this->keep_running)
	{
		centre_found = false;

		//Grab input frames.
		ps_frame	= NULL;
		ps_approx	= NULL;
		
		GETBUF(in_video, ps_frame, 100);

		if(SUCCESS != error || ps_frame == NULL) 
		{
			cout << "Error in reading input frame\n" << error ;
			continue;
		}

		if( ps_frame->empty() )
			continue;


		GETBUF(in_approx_pos, ps_approx, 100);

		if(SUCCESS != error || NULL == ps_approx) 
		{
			cout << "Error in reading segmented frame\n";
			delete ps_frame;
			ps_frame = NULL;
			continue;
		}


		//All input frames grabbed.

		/*************************************************************
		Cut out a region around the approximate position, size w x h.
		Run an edge detector on that region.
		Run circle-hough on that region.
		*************************************************************/
		
		//Make sure that the regions are within the matrix.
		if( ps_approx->y < h/2 )
		{
			cout << "Hough: Approximate region is too high\n";
			ps_approx->y = h/2;
		}
		if( ps_approx->y > ps_frame->rows - h/2)
		{
			cout << "Hough: Approximate region is too low\n";
			ps_approx->y = ps_frame->rows - h/2;
		}
		if( ps_approx->x < w/2)
		{
			cout << "Hough: Approximate x is too much to the left\n";
			ps_approx->x = w/2;
		}
		if( ps_approx->x > ps_frame->cols - w/2)
		{
			cout << "Hough: Approximate region is too much to the right\n";
			ps_approx->x = ps_frame->cols - w/2;
		}


		//Cut out a region.
		//Upper left of eye window
		eyew_ul.x = ps_approx->x-w/2;
		eyew_ul.y = ps_approx->y-h/2;
		//Bottom right
		eyew_br.x = ps_approx->x + w/2;
		eyew_br.y = ps_approx->y + h/2;
		
		
		Mat roi_rows = ps_frame->rowRange(eyew_ul.y, eyew_br.y);
		Mat eye_region = roi_rows.colRange(eyew_ul.x,  eyew_br.x);
		 

		//Smoothen it.
		GaussianBlur(eye_region, eye_region, Size(7,7), 2, 2);
		Mat eye_threshed;// (eye_region);

		//Threshold it - seems to work effectively to remove noise.
		threshold(eye_region, eye_threshed, (double)thresh, 255, THRESH_TRUNC);

		imshow("Edges", eye_threshed);
		waitKey(1);

		if( !dummy)
		{
			createTrackbar( "Threshold",
							"Edges",
							&thresh, 
							250,
							NULL,
							NULL);
		}
		thresh = getTrackbarPos("Threshold", "Edges");

		//Run the Hough transform. The hough transform directly works on a grey-scale image,
		//so there is no need to run an edge operator!
		vector<Vec3f> circles;
		HoughCircles(eye_threshed,		//input
					 circles,			//Output
					 CV_HOUGH_GRADIENT, //Type of transform
					 3,					//Accumulator ratio
					 10, //Minimum distance between circle centres
					 canny_param, //Some parameter for canny filter
					 hough_acc_thresh, //Threshold
					 min_radius, //Min radius //75
					 min_radius + 6); //Max radius

		//draw the circles.
		if(circles.size() > 0)
		{
			int i = 0;
			
			centre_found = true;
			//Centre of the circle. This is relative to the eye window.
			centre = Point(cvRound(circles[i][0]), cvRound(circles[i][1]));
			int radius = cvRound(circles[i][2]);

			if( this->last_centre.x != -1)
			{
				float new_alpha = (float) alpha / 100;
				//Smoothen out the relative centre.
				centre.x = new_alpha*this->last_centre.x + (1-new_alpha)*centre.x;
				centre.y = new_alpha*this->last_centre.y + (1-new_alpha)*centre.y;
				this->last_centre.x = centre.x;
				this->last_centre.y = centre.y;
			}
			else
			{
				//First time.
				this->last_centre = centre;
			}

			// draw the circle centre
			circle( eye_region, centre, 3, Scalar(0,255,0), -1);
			// draw the circle outline
			circle( eye_region, centre, radius, Scalar(255,255,255), 1);

			//Calculate actual centre, relative to the full frame.
			centre.x = centre.x + eyew_ul.x;
			centre.y = centre.y + eyew_ul.y;
		}

		imshow("cut-out", eye_region);
		
		if( !dummy)
		{
			createTrackbar( "Acc_thresh",
							"cut-out",
							&hough_acc_thresh, 
							20,
							NULL,
							NULL);

			createTrackbar( "Min_radius",
							"cut-out",
							&min_radius, 
							60,
							NULL,
							NULL);

			createTrackbar( "alpha",
							"cut-out",
							&alpha, 
							100,
							NULL,
							NULL);
			dummy = 1;
		}

		//Get the acc threshold from the trackbar
		hough_acc_thresh  = getTrackbarPos("Acc_thresh","cut-out"); 
		min_radius  = getTrackbarPos("Min_radius","cut-out"); 
//		max_radius  = getTrackbarPos("Max_radius","cut-out"); 
		//canny_param  = getTrackbarPos("canny_param","cut-out"); 
		alpha  = getTrackbarPos("alpha","cut-out"); 

		waitKey(1);
		
		//The eye frame and approx coordinates are no longer needed; delete them
		delete ps_frame;
		delete ps_approx;

		//Send this data out.
		if( this->out_coords && true == centre_found )
		{
			//Point *eye_center = new Point(this->last_centre);
			Point *eye_center = new Point(centre);

			error = this->out_coords->enqueue((void*)eye_center, 100);
			
			if( SUCCESS != error)
			{
				if( TIMEOUT != error)
					cout << "Hough: Unable to push out scene " << error << endl;
				delete eye_center;
			}
		}


		//Pull and push a scene frame.
		if( this->in_scene)
		{
			Mat *ps_scene;

			GETBUF(in_scene, ps_scene, 100);

			if( SUCCESS == error )
			{
				if( this->out_scene )
				{
					error = this->out_scene->enqueue(ps_scene, 100);
					if( SUCCESS != error)
					{
						if( TIMEOUT != error)
							cout << "Hough: Unable to push out scene " << error << endl;
						delete ps_scene;
					}
				}
				else
					delete ps_scene;
			}
			
		}

		
	}

	return SUCCESS;

}