/******************************************************************************
queue.cpp: Implementation of a FIFO queue. To be used to pass objects, usually 
		   Mat objects, between threads in the auto-eye-tracker system.
******************************************************************************/
#include <windows.h>
#include <iostream>

#include "common.h"
#include "queue.h"

using namespace std;

#define RETURN_TIMEOUT_OR_ERROR(E) {\
										if( E)\
										{\
											if(WAIT_TIMEOUT == E) return TIMEOUT;\
											else return MISC_ERROR;\
										}\
									}

/*******************************************************
Constructor for the queue class.
*******************************************************/
queue_t::queue_t()
{
	SECURITY_ATTRIBUTES security_attr = {sizeof(SECURITY_ATTRIBUTES), NULL, NULL};
	int i;

	//Initialize the counts.
	this->write = 0;
	this->read = 0;
	this->length = 0;

	this->space_present = CreateSemaphore(&security_attr, MAX_Q_LENGTH, MAX_Q_LENGTH, NULL /*(LPCWSTR)"semw"*/);

	if(NULL == this->space_present)
	{
		cout << "Q-create: Error in creating writer semaphore\n";
	}

	this->objects_present = CreateSemaphore(&security_attr, 0, MAX_Q_LENGTH, NULL /*(LPCWSTR)"semr"*/);

	if(NULL == this->objects_present)
	{
		cout << "Q-create: Error in creating reader semaphore\n";
	}

	for(i=0; i<MAX_Q_LENGTH; i++)
	{
		this->obj_array[i] = NULL;
	}

	this->lock = CreateMutex(NULL, FALSE, NULL);
	if(NULL == this->lock)
	{
		cout << "Q-create: Error in creating lock\n";
	}
}

/*******************************************************
Destructor for the queue class.
*******************************************************/
queue_t::~queue_t()
{
	int i;

	CloseHandle(this->space_present);
	CloseHandle(this->objects_present);
	
	for(i=0; i<MAX_Q_LENGTH; i++)
	{
		this->obj_array[i] = NULL;
	}

	//Initialize the counts.
	this->write = 0;
	this->read = 0;
	this->length = 0;
}


/*******************************************************
Enqueue funciton
timeout is in milliseconds.
*******************************************************/
status_t queue_t::enqueue(void *object, int timeout)
{
	int error1;
	int error;
	long temp;

	//Wait until there is some space.
	error1 = WaitForSingleObject(this->space_present, timeout);
	RETURN_TIMEOUT_OR_ERROR(error1);
	
	//Add this to the queue.

	//Lock the queue mutex to make sure that only one person gets in.
	error = WaitForSingleObject( this->lock, INFINITE ); 
	RETURN_TIMEOUT_OR_ERROR(error);
	
	this->obj_array[this->write] = object;
	this->write = (this->write+1)%MAX_Q_LENGTH;
	this->length++;
	
	//Release the lock
	ReleaseMutex(this->lock);

	//Signal the readers that there are objects available in the queue.
	error = ReleaseSemaphore(this->objects_present, 1, &temp);
	if(false == error)
	{
		return MISC_ERROR;
	}

	if( object == NULL)
	{
		cout << "pushing NULL?\n";
	}

	return SUCCESS;
}


/*******************************************************
Dequeue funciton
timeout is in milliseconds.
*******************************************************/
status_t queue_t::dequeue(void **object, int timeout)
{
	int error;
	int error1;
	long temp;

	//Wait until there is some space.
	error1 = WaitForSingleObject(this->objects_present, timeout);
	RETURN_TIMEOUT_OR_ERROR(error1);
	
	//Lock the queue mutex to make sure that only one person gets in.
	error = WaitForSingleObject( this->lock, INFINITE ); 
	RETURN_TIMEOUT_OR_ERROR(error);
	
	*object = this->obj_array[this->read];
	this->read = (read + 1)%MAX_Q_LENGTH;
	this->length--;
	
	
	//Release the lock
	ReleaseMutex(this->lock);

	//Signal the readers that there is some space present.
	error = ReleaseSemaphore(this->space_present, 1, &temp);
	if(false == error)
	{
		return MISC_ERROR;
	}

	if( *object == NULL)
	{
		cout << "pop NULL?\n";
	}

	return SUCCESS;
}