
/*****************************************************************
Video capture class.
Implements video capture using the OpenCV interface.
*****************************************************************/

#include <windows.h>
#include <iostream>
#include "cv.h"
#include "cvaux.h"
#include "highgui.h"
#include "common.h"
#include "queue.h"

using namespace cv;
using namespace std;

#include "video_capture.h"




//Thread function. Strictly a dummy placeholder!
static DWORD WINAPI thread(LPVOID pv_params)
{
	video_capture	*vcap = (video_capture*)pv_params;

	vcap->process();

	return 0;
}


/****************************************************************
Constructor for the video capture class.
****************************************************************/
video_capture::video_capture(int camera_id, int w, int h)
{
	SECURITY_ATTRIBUTES security_attr = {sizeof(SECURITY_ATTRIBUTES), NULL, NULL};

	this->output_q = NULL;

	this->camera_id = camera_id;
	this->keep_running = true;
	this->h_cap = NULL;
	this->h_thread = NULL;
	this->start_sem = NULL;
	this->h = h;
	this->w = w;

	//Semaphore to signal start of capture.
	//this->start_sem = CreateSemaphore(&security_attr, 1, 0, (LPCWSTR)"semw");
	this->start_sem = CreateSemaphore(NULL, 0, 1, NULL);

	if( NULL == this->start_sem)
	{
		cout << "Video capture: Error in creating semaphore\n";
		return;
	}


	//Open the camera.
	this->h_cap = new VideoCapture(camera_id);

	if(NULL == this->h_cap)
	{
		cout << "Error in creating video capture\n";
		return;
	}

	if( !this->h_cap->isOpened())
	{
		cout << "Error in opening video capture\n";
		return;
	}

	//Launch the video capture thread.
	this->h_thread = CreateThread(	NULL,                   // default security attributes
									0,                      // use default stack size  
									thread,       // thread function name
									this,          // argument to thread function 
									0,                      // use default creation flags 
									&this->thread_id);   // returns the thread identifier 

	if(NULL == this->h_thread)
	{
		cout << "Error in creating video capture thread\n";
		return;
	}

	cout << "Video camera "<< this->camera_id << " has started.\n";
}

/****************************************************************
Destructor for the video capture class.
****************************************************************/
video_capture::~video_capture(void)
{
	//Signal the thread to close.
	this->keep_running = false;

	//Post on the semaphore, in case it has not been started.
	this->start();

	// Wait for it to finish.
	WaitForSingleObject(this->h_thread, INFINITE);

	//Close the camera.
	delete this->h_cap;

	//Free the semaphore.
	CloseHandle(this->start_sem);

	//That is it!
	cout << "Video camera "<< this->camera_id << " stopped.\n";
}

status_t video_capture::set_output_queue(queue_t *q)
{
	this->output_q = q;

	return SUCCESS;
}

//Signal the video capture thread to actually start the capture.
status_t video_capture::start()
{
	LONG dummy;
	
	//Post on the semaphore.
	ReleaseSemaphore(this->start_sem, 1, &dummy);

	return SUCCESS;
}

/******************************************************
Actual video capture thread.

Capture new frames, and pushed them out into a queue.
******************************************************/
status_t video_capture::process()
{
	status_t error;
	Mat *grey_frame;
	Mat captured_frame;
	int capture_time;
	int elapsed;
	
	//Wait for the start signal from the main function.
	WaitForSingleObject(this->start_sem, INFINITE);

	char window_name[50];

	sprintf(window_name, "Camera_%d", this->camera_id);

	for (int j = 0; j < 10; j++)
	{
		(*this->h_cap)>>captured_frame;
		waitKey(30);
	}


	while(this->keep_running)
	{
		//Create a new frame, and capture into it. This will be released at the end.
		grey_frame = new Mat;
		
		//Capture!
		(*this->h_cap)>>captured_frame;
		capture_time = clock();

		if( captured_frame.cols > MAX_COLS_POSSIBLE ||
			captured_frame.cols <= 0 ||
			captured_frame.rows <= 0 ||
			captured_frame.rows > MAX_ROWS_POSSIBLE )
		{
			//Corrupt frame!
			waitKey(30);
			cout << "Video capture " << this->camera_id << ": Corrupt frame\n";
			delete grey_frame;
			grey_frame = NULL;
			continue;
		}

		//Convert to greyscale
		cvtColor(captured_frame, *grey_frame, CV_RGB2GRAY);

		//Add a gaussian blur
		GaussianBlur(*grey_frame, *grey_frame, Size(21,21), 1.5, 1.5);

		//Paint a rectangle showing the bounding box.
		if(grey_frame && this->h > 0 )
		{
			Point lowleft(w/2, grey_frame->rows - h/2);
			Point highright(grey_frame->cols - w/2, h/2);

			//Draw the bounding box.
			rectangle(*grey_frame, lowleft, highright, Scalar(0,0,0));
		}
		//Show the image.
		imshow(window_name, *grey_frame);


		//Push the frame out to the next node.
		if( this->output_q )
		{
			error = this->output_q->enqueue( (void*)grey_frame, 100);

			if(SUCCESS != error)
			{
				if(TIMEOUT != error)
					cout << "Vcap " << camera_id << ": Push output buffer failed\n";
				//Free the last allocated matrix, as it is not going to be freed further down the line
				delete grey_frame; 
				grey_frame = NULL;
			}
		}
		else
		{
			delete grey_frame;
			grey_frame = NULL;

		}

		

		elapsed = clock() - capture_time;

		if( elapsed < FRAME_PERIOD)
		{
			waitKey(FRAME_PERIOD - elapsed);
		}
		else
			waitKey(1);
	}

	return SUCCESS;
}